/*
 * Copyright (c) 1999 Mike Muir
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "chm.h"

static int io_file;
static int open_smb;

int OpenIO();
int ReadByte(u_char *return_value, int addr);
int WriteByte(int addr, int value);

const double vfactor[]={1,1,1,1.67,4,-4,-1.67};
const char *v_name[]={"VCore","Vit","Vio","+5V","+12V","-12V","-5V"};

int ReadRegister(int reg);

int ReadRegister(int reg) {
	u_char byte_read;
	
	ReadByte(&byte_read, reg);
	printf("Register 0x%x Reading: 0x%x\n", reg, byte_read);
	exit(1);
}

int WriteRegister(int reg, int val);

int WriteRegister(int reg, int val) {
	u_char byte_read;
	u_char byte_read_post;	

	ReadByte(&byte_read, reg);
	WriteByte(reg, val);
	ReadByte(&byte_read_post, reg);
        printf("Register 0x%x Was: 0x%x Now: 0x%x\n", reg, byte_read, byte_read_post);
	exit(1);
}

int SetDuty(int duty);

int SetDuty(int duty) {
	int duty_cent;
	WriteByte(0x5a, duty);
	WriteByte(0x5b, duty);
	duty_cent = (int)((duty *100) / 255);
	printf("Fans 1 and 2 set to %d%%\n", duty_cent);
	exit(1);
}

void usage();

void usage() {
	printf("\nConsole Hardware Monitor %s\n\n",CHM_VERSION);
 	(void)fprintf(stderr, "%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
	"usage: chm [options]\n",
	"	-I | -S			/dev/io method | SMBUS method (if supported)",
	"	-f      		Use Fahrenheit temperature scale",
	"	-k			Use Kelvin temperature scale",
	"	-x %			Set Fan Duty Cycle to %",
	"	-r d			Read Register d (decimal) from Chipset",
	"	-w v			Write value v to register d (used with -r)",
	"				to chipset.",
	"					***KNOW WHAT YOU ARE DOING***",
	"	-h      		Display this help screen",
	"	-d #			Number of microseconds to wait between",
	"				updates [default: 1000000] [use 0 for no loop]",
	" ",
	"	NOTE:	-x is experimental and is only known to work with the WinBond",
	"  		W83782D Chipset, as found on the ABIT BP6.\n");
	exit(1);
}

int main (int argc, char **argv)
{
	short scale = CELCIUS;
	char temp_char = 'C';
	int delay = 1000000, mb_temp, cpu0_temp, cpu1_temp, i, sw_char, fandiv[3];
	unsigned char smb_return=0;
	unsigned char cpu0_return=0;
	unsigned char cpu1_return=0;
	unsigned char mb_return=0;

	u_char return_value;
	u_char return_value2;
	int file_handle;
	u_char reg;
	u_char duty;
	u_char val = 0;
	u_char io_write = 0;

// Command line args

	while ((sw_char = getopt(argc, argv, "x: w: r: D: e: I S f k h d:")) != -1)
		switch (sw_char) {
	   case 'I':
         printf("Using ISA Method.\n");
         SMBUS = 0;
         break;
      case 'S':
         #ifdef HAVE_SMBUS
         printf("Using SMBUS Method.\n");
         SMBUS = 1;
         #else
         printf("Not compiled for use with SMBUS, reverting to ISA Method.\n");
         SMBUS = 0;
         #endif
         break;
		case 'w':
			val = atoi(optarg);
			io_write = 1;
			break;
		case 'r':
			reg = atoi(optarg);
			OpenIO();
			if (io_write)
            WriteRegister(reg, val);
			else 
				ReadRegister(reg);
			break;
		case 'x':
			duty = atoi(optarg);
			duty = (int)duty*2.55;
			duty += 1;
			OpenIO();
			printf("Setting Fan Duty to %d/255\n",duty);
			SetDuty(duty);
			break;
		case 'f':
			printf("Using Fahrenheit scale. \n\n");
			scale = FAHRENHEIT;
			break;
		case 'k':
			printf("Using Kelvin scale. \n\n");
			scale = KELVIN;
			break;
		case 'h':
			usage();
			break;
		case 'd':
			delay = atoi(optarg);
			printf("Delay: %d microseconds. \n\n",delay);
			break;
		default:
	}
	
	file_handle = OpenIO();

	printf("----------------------------------------\n");
	
        {
                static const int fandivide[]={1,2,4,8};
        
                ReadByte(&return_value, 0x47);
                smb_return = return_value;

                fandiv[0]=fandivide[(smb_return>>4)&3];
                fandiv[1]=fandivide[(smb_return>>6)&3];
                fandiv[2]=2;
        }

	while (1) {

// Motherboard Temperature (LM79/78 onchip thermistors)

      ReadByte(&return_value, 0x27);
#ifdef HAVE_BP6
	   cpu1_return = return_value;
#else
      mb_return = return_value;
#endif
// CPU_0 Temperature
				WriteByte(0x4E,0x01);
                                ReadByte(&return_value, 0x50);
				ReadByte(&return_value2, 0x51);
				if (return_value2 == 128) return_value += 0.5;
				WriteByte(0x4E,0);
				if (return_value > 200) return_value = 0.0;
                                cpu0_return = return_value;
 
// CPU_1 Temperature

				WriteByte(0x4E,0x02);
                                ReadByte(&return_value, 0x50);
				ReadByte(&return_value2, 0x51);
				if (return_value2 == 128) return_value += 0.5;
				WriteByte(0x4E,0);
				if (return_value > 200) return_value = 0.0;
#ifdef HAVE_BP6
	mb_return = return_value;
#else
				cpu1_return = return_value;
#endif

		switch (scale) {
			case FAHRENHEIT:
				mb_temp = (int)(mb_return*1.8) + 32;
				cpu0_temp = (int)(cpu0_return*1.8) + 32;
				cpu1_temp = (int)(cpu1_return*1.8) + 32;
				temp_char = 'F';
				break;
			case KELVIN:
				mb_temp = (int)(mb_return+273);	
				cpu0_temp = (int)(cpu0_return+273);
				cpu1_temp = (int)(cpu1_return+273);
				temp_char = 'K';	
				break;
			default:
				mb_temp = (int)mb_return;
				cpu0_temp = (int)cpu0_return;
				cpu1_temp = (int)cpu1_return;
				temp_char = 'C';
		}

		printf("Motherboard Temperature: %d \260 %c\n",mb_temp,temp_char);
		printf("CPU_0 Temperature: %d \260 %c\n",cpu0_temp,temp_char);
		printf("CPU_1 Temperature: %d \260 %c\n",cpu1_temp,temp_char);

// Voltages

		for (i=0x20;i<0x27;i++) {
			
                	ReadByte(&return_value, i);
                	smb_return = return_value;

			printf("%s: \t %g V \n",v_name[i-0x20],smb_return*vfactor[i-0x20]/64);
		}	

// Fan Guages

		for(i=0x28;i<0x2b;i++){

                        ReadByte(&return_value, i);
                        smb_return = return_value;

			if(smb_return!=0xff){
				printf("Fan %d: %d rpm\n",(i-0x28)+1,(int)(1.35E6/(smb_return*fandiv[i-0x28])));
			} else {
				printf("Fan %d: Not Available\n",(i-0x28)+1);
			}
		}
		if (delay != 0) {
			printf("----------------------------------------\n");
			usleep(delay);
		} else {
			exit(1);
		}
	} 	
}

int OpenIO()
{
	#ifdef HAVE_SMBUS
		if (SMBUS)
		{
		u_char smb_return;
		struct smbcmd cmd;
		cmd.slave=0x5a;
		cmd.data.byte_ptr=&smb_return;
		if((open_smb=open("/dev/smb0",000))<0){
			fprintf(stderr, "Failed to open /dev/smb0.\n");
			exit (-1);
		}
		cmd.cmd=0x47;
		if(ioctl(open_smb,SMB_READB,&cmd)==-1){
		  perror("ioctl");
		  return -1;
		}
		return 0;
		} else {
			if((io_file=open("/dev/io",000))<0)
        		{
                		fprintf(stderr, "Failed to open /dev/io.\n");
        			exit (-1);
        		}
        		return 0;
		}
	#else
        if((io_file=open("/dev/io",000))<0)
        {
                fprintf(stderr, "Failed to open /dev/io.\n");
                exit (1);
        }
        return 0;
	#endif
}

int ReadByte(u_char *return_value, int addr)
{
		#ifdef HAVE_SMBUS
                        if (SMBUS) {
				u_char smb_return;
				struct smbcmd cmd;
				cmd.slave=0x5a;
				cmd.data.byte_ptr=&smb_return;
                                cmd.cmd=addr;
                                if(ioctl(open_smb,SMB_READB,&cmd)==-1){
                                        perror("IOCTL");
                                        return -1;
                                }
				*return_value = smb_return;
				return 0;
                        } else {
        			outb(WBIO1,addr);
				*return_value = inb(WBIO2);
				return 0;
                        }
                #else 
   	   		outb(WBIO1,addr);
 	      		*return_value = inb(WBIO2);
	        	return 0;
                #endif
}

int WriteByte(int addr, int value)
{
	#ifdef HAVE_SMBUS
	if (SMBUS) {
		u_char smb_return;
		struct smbcmd cmd;
		cmd.slave=0x5a;
		cmd.data.byte_ptr=&smb_return;
		cmd.cmd=addr;
		cmd.data.byte=value;
		if(ioctl(open_smb,SMB_WRITEB,&cmd)==-1){
			perror("IOCTL");
			return -1;
		}
	} else {
                outb(WBIO1,addr);
		outb(WBIO2,value);
        }
        #else 
                outb(WBIO1,addr);
		outb(WBIO2,value);
        #endif
	return 0;
}
